from typing import Union

import numpy as np
from sklearn.metrics import accuracy_score


def safe_accuracy_score(y_true: np.ndarray, y_pred: np.ndarray, **kwargs) -> float:
    """
    Wrapper around sklearn accuracy store that returns zero for empty sequences of labels

    :param y_true: Ground truth (correct) labels.
    :param y_pred: Predicted labels, as returned by a classifier.
    :param kwargs:
    :return:
    """
    if len(y_true) == len(y_pred) == 0:
        return 0
    return accuracy_score(y_true, y_pred, **kwargs)


def in_simplex(probabilities: np.ndarray, num_classes=None) -> bool:
    """

    :param probabilities: single vector of probabilities of shape (n_classes,) or multiple
        vectors as array of shape (n_samples, n_classes)
    :param num_classes: if provided, will check whether probability vectors have the correct number of classes
    :return:
    """
    if len(probabilities.shape) == 1:
        probabilities = probabilities[None, :]
    if num_classes is None:
        num_classes = probabilities.shape[1]

    return (
        probabilities.shape[1] == num_classes
        and np.allclose(np.sum(probabilities, axis=1), 1.0, rtol=0.01)
        and (probabilities >= 0).all()
        and (probabilities <= 1).all()
    )


def sample_index(probabilities: np.ndarray) -> Union[int, np.ndarray]:
    """
    Sample indices with the input probabilities. This is essentially a vectorized
    version of np.random.choice

    :param probabilities: single vector of probabilities of shape (n_indices-1,) or multiple
        vectors as array of shape (n_samples, n_indices-1)
    :return: index or array of indices
    """
    rng = np.random.default_rng()
    if len(probabilities.shape) == 1:
        return rng.choice(len(probabilities), p=probabilities)
    elif len(probabilities.shape) == 2:
        # this is a vectorized implementation of np.random.choice with inverse transform sampling
        # see e.g. https://stephens999.github.io/fiveMinuteStats/inverse_transform_sampling.html
        # and https://stackoverflow.com/questions/47722005/vectorizing-numpy-random-choice-for-given-2d-array-of-probabilities-along-an-a
        random_uniform = rng.random(len(probabilities))[:, None]
        return (probabilities.cumsum(axis=1) > random_uniform).argmax(axis=1)
    else:
        raise ValueError(
            f"Unsupported input shape: {probabilities.shape}. Can only be 1 or 2 dimensional."
        )



from abc import ABC, abstractmethod
from typing import Generic, List, Optional, TypeVar

import netcal.binning as bn
import netcal.scaling as scl
import numpy as np
from netcal import AbstractCalibration
from sklearn.base import BaseEstimator


class BaseCalibrationMethod(ABC, BaseEstimator):
    @abstractmethod
    def fit(self, confidences: np.ndarray, ground_truth: np.ndarray):
        pass

    @abstractmethod
    def get_calibrated_confidences(self, confidences: np.ndarray):
        pass

    def __str__(self):
        return self.__class__.__name__


def _get_confidences_from_netcal_calibrator(
    confidences: np.ndarray, calibrator: AbstractCalibration
):
    calibrated_confs = calibrator.transform(confidences)

    # TODO: there is a whole bunch of hacks here. I want to get rid of netcal, don't like the code there
    # unfortunately, for 2-dim input netcal gives only the probabilities for the second class,
    # changing the dimension of the output array
    if calibrated_confs.ndim < 2:
        second_class_confs = calibrated_confs
        first_class_confs = 1 - second_class_confs
        calibrated_confs = np.stack([first_class_confs, second_class_confs], axis=1)

        if (
            len(confidences) == 1
        ):  # Netcal has a bug for single data points, this is a dirty fix
            calibrated_confs = calibrated_confs[None, 0]

        if calibrated_confs.shape != confidences.shape:
            raise RuntimeError(
                f"Shape mismatch for input {confidences}, output {calibrated_confs}. "
                f"Netcal output: {second_class_confs}"
            )

    return calibrated_confs


TNetcalModel = TypeVar("TNetcalModel", bound=AbstractCalibration)


# TODO: this is definitely not the final class structure. For now its ok, I want to completely decouple from netcal soon
class NetcalBasedCalibration(BaseCalibrationMethod, Generic[TNetcalModel]):
    def __init__(self, netcal_model: TNetcalModel):
        self.netcal_model = netcal_model

    def fit(self, confidences: np.ndarray, ground_truth: np.ndarray):
        self.netcal_model.fit(confidences, ground_truth)

    def get_calibrated_confidences(self, confidences: np.ndarray) -> np.ndarray:
        return _get_confidences_from_netcal_calibrator(confidences, self.netcal_model)


class TemperatureScaling(NetcalBasedCalibration[scl.TemperatureScaling]):
    def __init__(self):
        super().__init__(scl.TemperatureScaling())


class BetaCalibration(NetcalBasedCalibration[scl.BetaCalibration]):
    def __init__(self):
        super().__init__(scl.BetaCalibration())


class LogisticCalibration(NetcalBasedCalibration[scl.LogisticCalibration]):
    def __init__(self):
        super().__init__(scl.LogisticCalibration())


class IsotonicRegression(NetcalBasedCalibration[bn.IsotonicRegression]):
    def __init__(self):
        super().__init__(bn.IsotonicRegression())


class HistogramBinning(NetcalBasedCalibration[bn.HistogramBinning]):
    def __init__(self, bins=20):
        super().__init__(bn.HistogramBinning(bins=bins))
        self.bins = bins


class ClassWiseCalibration(BaseCalibrationMethod):
    def __init__(self, calibration_method_factory=TemperatureScaling):
        self.calibration_method_factory = calibration_method_factory
        self.n_classes: Optional[int] = None
        self.calibration_methods: Optional[List[BaseCalibrationMethod]] = None

    # TODO: maybe parallelize this and predict
    def fit(self, confidences: np.ndarray, labels: np.ndarray):
        self.n_classes = confidences.shape[1]
        self.calibration_methods = []
        for class_label in range(self.n_classes):
            calibration_method = self.calibration_method_factory()
            selected_confs, selected_labels = get_class_confs_labels(
                class_label, confidences, labels
            )
            calibration_method.fit(selected_confs, selected_labels)
            self.calibration_methods.append(calibration_method)

    def get_calibrated_confidences(self, confs: np.ndarray):
        result = np.zeros(confs.shape)
        argmax = confs.argmax(1)
        for class_label in range(self.n_classes):
            scaler = self.calibration_methods[class_label]
            indices = argmax == class_label
            selected_confs = confs[indices]
            calibrated_confs = scaler.get_calibrated_confidences(selected_confs)
            assert calibrated_confs.shape == selected_confs.shape, (
                f"Expected shape {selected_confs.shape} but got {calibrated_confs.shape}. Confs: "
                f"{selected_confs}, output: {calibrated_confs}"
            )

            result[indices] = calibrated_confs
        return result


class ConfidenceReducedCalibration(BaseCalibrationMethod, BaseEstimator):
    def __init__(self, calibration_method=TemperatureScaling()):
        self.calibration_method = calibration_method

    def fit(self, confidences: np.ndarray, ground_truth: np.ndarray):
        reduced_confs, reduced_gt = get_binary_classification_data(
            confidences, ground_truth
        )
        self.calibration_method.fit(reduced_confs, reduced_gt)

    def get_calibrated_confidences(self, confidences: np.ndarray):
        reduced_confs = get_reduced_confidences(confidences)
        reduced_predictions = self.calibration_method.get_calibrated_confidences(
            reduced_confs
        )
        reduced_predictions = reduced_predictions[:, 0]  # take only 0-class prediction
        n_classes = confidences.shape[1]
        non_predicted_class_confidences = (1 - reduced_predictions) / (n_classes - 1)

        # using broadcasting here
        calibrated_confidences = (
            non_predicted_class_confidences * np.ones(confidences.shape).T
        )
        calibrated_confidences = calibrated_confidences.T

        argmax_indices = np.expand_dims(confidences.argmax(axis=1), axis=1)
        np.put_along_axis(
            calibrated_confidences, argmax_indices, reduced_predictions[:, None], axis=1
        )
        assert np.all(
            np.isclose(calibrated_confidences.sum(1), np.ones(len(confidences)))
        )
        assert calibrated_confidences.shape == confidences.shape
        return calibrated_confidences


def get_class_confs_labels(c: int, confidences: np.ndarray, labels: np.ndarray):
    indices = confidences.argmax(1) == c
    return confidences[indices], labels[indices]


def get_reduced_confidences(confidences: np.ndarray):
    top_class_predictions = confidences.max(axis=1)
    return np.stack([top_class_predictions, 1 - top_class_predictions], axis=1)


def get_binary_classification_data(confidences: np.ndarray, labels: np.ndarray):
    new_confidences = get_reduced_confidences(confidences)
    pred_was_correct = labels == confidences.argmax(axis=1)
    # this is a hack - we predict class 0 if pred was correct, else class 1
    new_gt = (np.logical_not(pred_was_correct)).astype(int)
    return new_confidences, new_gt



import logging
from typing import Literal, NamedTuple, Protocol, Sequence, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import ListedColormap
from sklearn.utils import check_consistent_length, column_or_1d

log = logging.getLogger(__name__)


class ReliabilityResult(NamedTuple):
    prob_true: np.ndarray
    prob_pred: np.ndarray
    n_members: np.ndarray
    bin_edges: np.ndarray


def _assert_1d(arr: np.ndarray, arr_name: str, error_type=ValueError):
    if arr.ndim != 1:
        raise error_type(f"{arr_name} should be a 1d array but got shape: {arr.shape}")


def _to2d(confidences: np.ndarray):
    """
    If a 1d array is passed, we assume that it corresponds to the confidence in True class, i.e. in label=1.
    Then to get the confs in the right order, we prepend 1-confidences for the label 0
    """
    if confidences.ndim == 2:
        return confidences
    if confidences.ndim == 1:
        return np.stack([1 - confidences, confidences]).T
    raise ValueError(
        f"Cannot turn array of shape: {confidences.shape} to two-dim array."
    )


def _binary_classifier_reliability(
    y_true: np.ndarray,
    y_prob: np.ndarray,
    n_bins=12,
    strategy: Literal["uniform", "quantile"] = "uniform",
) -> ReliabilityResult:
    """
    The implementation is essentially a copy of `sklearn.calibration.calibration_curve` but contains additional
    quantities in the return: `n_members` and `bin_edges`. Unfortunately, the sklearn implementation cannot be
    used as it doesn't return these quantities, and they would have to be recomputed.
    Compute arrays related to the reliabilities of a binary classifiers. They can be used e.g. for computing
    calibration errors or for visualizing reliability curves.

    :param n_bins:
    :param y_true: the confidences in the prediction of the class **True**
        (or class 1, if labels passed as integers). Should be a numpy array of shape (n_samples, )
    :param y_prob: array of shape (n_samples, ) containing True and False or integers (1, 0) respectively
    :param strategy: Strategy used to define the widths of the bins.
        **uniform** The bins have identical widths.
        **quantile** The bins have the same number of samples and depend on y_prob.
    :return: named tuple containing arrays with confidences, accuracies, members, bin_edges
    """
    y_true = column_or_1d(y_true)
    y_prob = column_or_1d(y_prob)
    check_consistent_length(y_true, y_prob)

    if y_prob.min() < 0 or y_prob.max() > 1:
        raise ValueError("y_prob has values outside [0, 1].")

    uniform_bins = np.linspace(0.0, 1.0, n_bins + 1)
    if strategy == "quantile":  # Determine bin edges by distribution of data
        bins = np.quantile(y_prob, uniform_bins)
    elif strategy == "uniform":
        bins = uniform_bins
    else:
        raise ValueError(
            "Invalid entry to 'strategy' input. Strategy "
            "must be either 'quantile' or 'uniform'."
        )

    binids = np.searchsorted(bins[1:-1], y_prob)

    bin_sums = np.bincount(binids, weights=y_prob, minlength=len(bins))
    bin_true = np.bincount(binids, weights=y_true, minlength=len(bins))
    bin_total = np.bincount(binids, minlength=len(bins))

    is_nonempty = bin_total != 0
    n_empty_bins = n_bins - is_nonempty.sum()
    if n_empty_bins > 0:
        log.debug(
            f"{n_empty_bins} of {n_bins} bins were empty, the reliability curve cannot be estimated in them."
            f"This can be prevented by either: \n"
            f"  1) reducing the number of bins (current value is {n_bins}) or \n"
            f"  2) increasing the sample size (current value is {len(y_true)}) or \n"
            f"  3) using strategy='quantile'"
        )

    last_nonempty_bin = np.where(is_nonempty)[0][-1]
    if last_nonempty_bin == n_bins:
        last_bin_edge = 1.0
    else:
        last_bin_edge = bins[last_nonempty_bin + 1]

    bin_edges = np.append(bins[is_nonempty], last_bin_edge)
    bin_members = bin_total[is_nonempty]
    prob_true = bin_true[is_nonempty] / bin_members
    prob_pred = bin_sums[is_nonempty] / bin_members

    return ReliabilityResult(prob_true, prob_pred, bin_members, bin_edges)


def classifier_reliability(
    y_true: np.ndarray,
    confidences: np.ndarray,
    class_label: Union[int, Literal["top_class"]] = 0,
    n_bins=12,
    strategy: Literal["uniform", "quantile"] = "uniform",
):
    """
    Computes arrays related to the reliabilities of the provided confidences. They can be used e.g. for computing
    calibration errors or for visualizing reliability curves.

    :param y_true:
    :param confidences:
    :param n_bins:
    :param class_label: either an integer label for the class-wise reliabilities, or "top_class" for the
        reliabilities in predictions.
    :param strategy:

    :return: named tuple containing arrays with confidences, accuracies, members, bin_edges
    """
    confidences = _to2d(confidences)

    y_pred = confidences.argmax(axis=1)
    # noinspection PyArgumentList
    if class_label == "top_class":
        reduced_y_prob = np.take_along_axis(confidences, y_pred[:, None], axis=1).T[0]
        reduced_y_true = y_true == y_pred
    else:
        reduced_y_prob = confidences[:, class_label]
        reduced_y_true = y_true == class_label
    return _binary_classifier_reliability(
        reduced_y_true,
        reduced_y_prob,
        n_bins=n_bins,
        strategy=strategy,
    )


def expected_calibration_error(
    y_true: np.ndarray,
    confidences: np.ndarray,
    class_label: Union[int, Literal["top_class"]] = "top_class",
    n_bins=12,
    strategy: Literal["uniform", "quantile"] = "uniform",
):
    """
    :param class_label: if "top_class", will be the usual (confidence) ECE. Otherwise, it will be the
        marginal class-wise ECE for the selected class.
    :param n_bins:
    :param strategy:
    :return:
    """
    reliabilities = classifier_reliability(
        y_true,
        confidences,
        class_label=class_label,
        n_bins=n_bins,
        strategy=strategy,
    )
    sum_members = np.sum(reliabilities.n_members)
    if sum_members == 0:
        return 0.0
    weights = reliabilities.n_members / sum_members
    abs_diff = np.abs(reliabilities.prob_pred - reliabilities.prob_true)
    return np.dot(abs_diff, weights)


def average_calibration_error(
    y_true: np.ndarray,
    confidences: np.ndarray,
    n_bins=12,
    strategy: Literal["uniform", "quantile"] = "uniform",
):
    reliabilities = classifier_reliability(
        y_true, confidences, class_label="top_class", n_bins=n_bins, strategy=strategy
    )
    abs_distances = np.abs(reliabilities.prob_pred - reliabilities.prob_true)
    return np.mean(abs_distances)


def max_calibration_error(
    y_true: np.ndarray,
    confidences: np.ndarray,
    n_bins=12,
    strategy: Literal["uniform", "quantile"] = "uniform",
):
    reliabilities = classifier_reliability(
        y_true, confidences, class_label="top_class", n_bins=n_bins, strategy=strategy
    )
    abs_distances = np.abs(reliabilities.prob_pred - reliabilities.prob_true)
    return np.max(abs_distances)


def class_wise_expected_calibration_error(
    y_true: np.ndarray,
    confidences: np.ndarray,
    n_bins=12,
    strategy: Literal["uniform", "quantile"] = "uniform",
):
    confidences = _to2d(confidences)
    num_classes = confidences.shape[-1]
    sum_marginal_errors = sum(
        expected_calibration_error(
            y_true, confidences, k, n_bins=n_bins, strategy=strategy
        )
        for k in range(num_classes)
    )
    return sum_marginal_errors / num_classes


class ReliabilitiesProviderProtocol(Protocol):
    def __call__(self, class_label: int, n_bins, strategy) -> ReliabilityResult:
        pass


def _plot_reliability_curves(
    reliabilities_provider: ReliabilitiesProviderProtocol,
    class_labels: Sequence[Union[int, Literal["top_class"]]],
    display_weights,
    n_bins,
    strategy: Literal["uniform", "quantile"],
):
    """
    Helper function to plot reliabilities. Within EvalStats part of the reliabilities
    is precomputed, and y_true and confidences are known - so we don't want to use the same
    provider there.

    :param reliabilities_provider:
    :param class_labels:
    :param display_weights:
    :param n_bins:
    :param strategy:
    :return:
    """
    colors = ListedColormap(["y", "g", "r", "c", "m"])

    fig = plt.figure()
    plt.title(f"Reliability curves ({n_bins} bins)")
    plt.xlabel("confidence")
    plt.ylabel("ground truth probability")
    plt.axis("equal")

    # plotting a diagonal for perfect calibration
    plt.plot([0, 1], [0, 1], label="perfect calibration", color="b")

    # for each class, plot curve and weights, cycle through colors
    for i, class_label in enumerate(class_labels):
        color = colors(i)
        if class_label == "top_class":
            plot_label = "prediction"
        else:
            plot_label = f"class {class_label}"

        prob_true, prob_pred, n_members, bin_edges = reliabilities_provider(
            class_label,
            n_bins=n_bins,
            strategy=strategy,
        )
        plt.plot(prob_pred, prob_true, marker=".", label=plot_label, color=color)
        if display_weights:
            # rescale the weights for improved visibility
            weights = n_members / (3 * n_members.max())
            width = np.diff(bin_edges)
            plt.bar(
                bin_edges[:-1],
                weights,
                align="edge",
                alpha=0.2,
                width=width,
                color=color,
                label=f"weights ({plot_label})",
                edgecolor="black",
                linewidth=0.5,
                linestyle="--",
            )

    axes = plt.gca()
    axes.set_xlim([0, 1])
    axes.set_ylim([0, 1])
    plt.legend(loc="best")
    return fig


def plot_reliability_curves(
    y_true: np.ndarray,
    confidences: np.ndarray,
    class_labels: Sequence[Union[int, Literal["top_class"]]],
    display_weights=False,
    n_bins=12,
    strategy: Literal["uniform", "quantile"] = "uniform",
):
    """
    :param y_true:
    :param confidences:
    :param class_labels:
    :param display_weights: If True, for each reliability curve the weights of each bin will be
        plotted as histogram. The weights have been scaled for the sake of display, only relative differences
        between them have an interpretable meaning.
        The errors containing "expected" in the name take these weights into account.
    :param strategy:
    :param n_bins:
    :return: figure
    """

    def reliabilities_provider(
        class_label,
        n_bins,
        strategy,
    ):
        return classifier_reliability(
            y_true,
            confidences,
            class_label,
            n_bins=n_bins,
            strategy=strategy,
        )

    return _plot_reliability_curves(
        reliabilities_provider, class_labels, display_weights, n_bins, strategy
    )



from abc import ABC, abstractmethod
from typing import Sequence, Union

import numpy as np
import scipy.optimize
import scipy.stats

from kyle.transformations import IdentitySimplexAut, SimplexAut
from kyle.util import sample_index


class FakeClassifier(ABC):
    def __init__(
        self,
        num_classes: int,
        simplex_automorphism: SimplexAut = None,
        check_io=True,
    ):
        if num_classes < 1:
            raise ValueError(f"{self.__class__.__name__} requires at least two classes")
        self.num_classes = num_classes
        self._rng = np.random.default_rng()

        self._simplex_automorphism: SimplexAut = None
        self.set_simplex_automorphism(simplex_automorphism)
        self.check_io = check_io

    # TODO or not TODO: one could get rid of separate SimplexAut. class in favor of passing a function
    #   pro: the function is less verbose to write, easier for user; contra: naming and state become more convoluted
    def set_simplex_automorphism(self, aut: Union[SimplexAut, None]) -> None:
        """
        :param aut: if None, the identity automorphism will be set
        """
        if aut is None:
            aut = IdentitySimplexAut(self.num_classes)
        if aut.num_classes is not None and aut.num_classes != self.num_classes:
            raise ValueError(f"{aut} has wrong number of classes: {aut.num_classes}")
        self._simplex_automorphism = aut

    @abstractmethod
    def sample_confidences(self, n_samples: int) -> np.ndarray:
        ...

    @property
    def simplex_automorphism(self):
        return self._simplex_automorphism

    def get_sample_arrays(self, n_samples: int):
        """
        Get arrays with ground truth and predicted probabilities

        :param n_samples:
        :return: tuple of arrays of shapes (n_samples,), (n_samples, n_classes)
        """
        calibrated_confidences = self.sample_confidences(n_samples)
        gt_labels = sample_index(calibrated_confidences)
        confidences = self.simplex_automorphism.transform(calibrated_confidences)
        return gt_labels, confidences

    def __str__(self):
        return f"{self.__class__.__name__}_{self.simplex_automorphism}"


class DirichletFC(FakeClassifier):
    def __init__(
        self,
        num_classes: int,
        alpha: Sequence[float] = None,
        simplex_automorphism: SimplexAut = None,
    ):
        super().__init__(num_classes, simplex_automorphism=simplex_automorphism)

        self._alpha: np.ndarray = None
        self.set_alpha(alpha)

    def set_alpha(self, alpha: Union[np.ndarray, None]):
        """
        :param alpha: if None, the default value of [1, ..., 1] will be set
        """
        if alpha is None:
            alpha = np.ones(self.num_classes)
        else:
            alpha = np.array(alpha)
            if not alpha.shape == (self.num_classes,):
                raise ValueError(f"Wrong shape of alpha: {alpha.shape}")
        self._alpha = alpha

    @property
    def alpha(self):
        return self._alpha

    def sample_confidences(self, n_samples: int) -> np.ndarray:
        return self._rng.dirichlet(self.alpha, size=n_samples)

    def pdf(self, confidences, alpha=None):
        if alpha is None:
            alpha = self.alpha
        return scipy.stats.dirichlet.pdf(confidences.T, alpha)

    def fit(self, confidences, initial_alpha=None, alpha_bounds=None, **kwargs):
        """
        Fits the dirichlet fake classifier to the provided confidence distribution using maximum likelihood estimation
        and sets the fake classifier parameters to the best fit parameters

        :param confidences: Numpy array of shape (num_samples, num_classes);
                            confidence distribution to fit classifier to
        :param initial_alpha: Float; Initial guess for fitting alpha parameters
        :param alpha_bounds: Tuple, (lower_bound, upper_bound); Bounds for fitting alpha parameters. A lower/upper bound
                             of None corresponds to unbounded parameter
        :param kwargs: passed to ``scipy.optimize.minimize``
        :return:
        """
        if initial_alpha is None:
            initial_alpha = self.alpha

        if alpha_bounds is None:
            alpha_bounds = (0.0001, None)

        # rescale confidences to avoid divergences on sides of simplex and renormalize
        confidences = (
            confidences * (confidences.shape[0] - 1) + 1 / self.num_classes
        ) / confidences.shape[0]
        confidences = confidences / np.sum(confidences, axis=1)[:, None]

        alpha_bounds = [alpha_bounds] * self.num_classes

        nll = lambda parm: -np.sum(np.log(self.pdf(confidences, parm)))
        mle_fit = scipy.optimize.minimize(
            nll, initial_alpha, bounds=alpha_bounds, **kwargs
        )
        self.set_alpha(mle_fit.x)

        return mle_fit


class MultiDirichletFC(FakeClassifier):
    """
    A fake classifier that first draws from a K categorical distribution and based on the result then draws from
    1 of K Dirichlet Distributions of a restricted form.
    The K'th Dirichlet Distribution has parameters of the form:  sigma * {1, 1, ..., alpha_k, 1, 1, ...}; alpha > 1
    where 'alpha_k' is at the k'th position.
    Effectively a distribution with a maximum of variable position and variable variance in each corner of the simplex

    :param num_classes:
    :param alpha: numpy array of shape (num_classes,). k'th entry corresponds to alpha_k for the k'th dirichlet
    :param sigma: numpy array of shape (num_classes,). k'th entry corresponds to sigma for the k'th dirichlet
    :param distribution_weights: numpy array of shape (num_classes,). Probabilities used for drawing from K Categorical
    :param simplex_automorphism:
    """

    def __init__(
        self,
        num_classes: int,
        alpha: Sequence[float] = None,
        sigma: Sequence[float] = None,
        distribution_weights: Sequence[float] = None,
        simplex_automorphism: SimplexAut = None,
    ):
        super().__init__(num_classes, simplex_automorphism=simplex_automorphism)

        self._alpha: np.ndarray = None
        self._sigma: np.ndarray = None
        self._distribution_weights: np.ndarray = None

        self.set_alpha(alpha)
        self.set_sigma(sigma)
        self.set_distribution_weights(distribution_weights)

    @property
    def alpha(self):
        return self._alpha

    def set_alpha(self, alpha: Union[np.ndarray, None]):
        """
        :param alpha: if None, the default value of [1, ..., 1] will be set.
        """
        if alpha is None:
            alpha = np.ones(self.num_classes)
        else:
            alpha = np.array(alpha)
            if not alpha.shape == (self.num_classes,):
                raise ValueError(f"Wrong shape of alpha: {alpha.shape}")
        self._alpha = alpha

    @property
    def sigma(self):
        return self._sigma

    def set_sigma(self, sigma: Union[np.ndarray, None]):
        """
        :param sigma: if None, the default value of [1, ..., 1] will be set
        """
        if sigma is None:
            sigma = np.ones(self.num_classes)
        else:
            sigma = np.array(sigma)
            if not sigma.shape == (self.num_classes,):
                raise ValueError(f"Wrong shape of sigma: {sigma.shape}")
        self._sigma = sigma

    @property
    def distribution_weights(self):
        return self._distribution_weights

    def set_distribution_weights(self, distribution_weights: Union[np.ndarray, None]):
        """
        :param distribution_weights: if None, the default value of [1/num_classes, ..., 1/num_classes] will be set
        """
        if distribution_weights is None:
            distribution_weights = np.ones(self.num_classes) / self.num_classes
        else:
            distribution_weights = np.array(distribution_weights)
            if not distribution_weights.shape == (self.num_classes,):
                raise ValueError(
                    f"Wrong shape of predicted_class_weights: {distribution_weights.shape}"
                )
        self._distribution_weights = distribution_weights / np.sum(distribution_weights)

    def get_parameters(self):
        return self._alpha, self._sigma, self._distribution_weights

    def set_parameters(self, alpha, sigma, distribution_weights):
        self.set_alpha(alpha)
        self.set_sigma(sigma)
        self.set_distribution_weights(distribution_weights)

    def sample_confidences(self, n_samples: int) -> np.ndarray:

        weight_array = np.repeat(self.distribution_weights[None, :], n_samples, axis=0)
        chosen_distributions = sample_index(weight_array)

        confidences = np.zeros((n_samples, self.num_classes))

        for i, chosen_distribution in enumerate(chosen_distributions):
            alpha_vector = np.ones(self.num_classes)
            alpha_vector[chosen_distribution] = self.alpha[chosen_distribution]
            alpha_vector *= self.sigma[chosen_distribution]

            confidences[i, :] = self._rng.dirichlet(alpha_vector)

        return confidences

    def pdf(self, confidences, alpha=None, sigma=None, distribution_weights=None):
        """
        Computes pdf of MultiDirichletFC. Using K categorical distribution to sample from K dirichlet distributions
        is equivalent to sampling from a pdf that is a weighted sum of the K individual dirichlet pdf's
        :param confidences: numpy array of shape (num_classes,) or (num_samples, num_classes)
        :param distribution_weights: numpy array of shape (num_classes,) uses self.distribution_weights if not provided
        :param sigma: numpy array of shape (num_classes,) uses self.sigma if not provided
        :param alpha: numpy array of shape (num_classes,) uses self.alpha if not provided
        """

        if alpha is None:
            alpha = self.alpha
        if sigma is None:
            sigma = self.sigma
        if distribution_weights is None:
            distribution_weights = self.distribution_weights

        confidences = confidences.T

        distributions = np.zeros(confidences.shape)

        for i, (a, s) in enumerate(zip(alpha, sigma)):
            alpha_vector = np.ones(self.num_classes)
            alpha_vector[i] = a
            alpha_vector *= s

            distributions[i] = scipy.stats.dirichlet.pdf(confidences, alpha_vector)

        return np.sum(distribution_weights[:, None] * distributions, axis=0) / np.sum(
            distribution_weights
        )

    def fit(
        self,
        confidences,
        initial_parameters=None,
        parameter_bounds=None,
        simplified_fitting=True,
        **kwargs,
    ):
        """
        Fits a Multi-Dirichlet fake classifier to the provided confidence distribution using maximum likeihood
        estimation and sets the fake classifier parameters to the best fit parameters.
        If simplified_fitting is set to False all parameters of the fake classifier are fit directly via MLE
        If simplified_fitting is set to True each dirichlet is fit separately. Alpha and Sigma of the k'th dirichlet
        are fit to the subset of the confidences that predict the k'th class, i.e. for which argmax(c) = k. The
        distribution weights are not fit, but estimated from the predicted class probabilities of the confidence
        distribution.

        :param confidences: Numpy array of shape (num_samples, num_classes);
                            confidence distribution to fit classifier to
        :param initial_parameters: Numpy array of shape (3,) ((2,) for simplified_fitting=True)
                       Corresponds to initial guesses for each parameter 'class' alpha, sigma and distribution_weights
                       If None, uses [1, 1, 1/num_classes]
        :param parameter_bounds: Sequence of 3 (2 for simplified_fitting=True) tuples (lower_bound, upper_bound)
                        Corresponds to the bounds on each parameter 'class',  alpha, sigma and distribution_weights
                        A lower/upper bound of None corresponds to unbounded parameters
                        If None, uses intervals [(0, + infinity), (0, + infinity), (0,1)]
        :param simplified_fitting: If False directly fits Multi-Dirichlet FC to confidence distribution
                                   If True fits each dirichlet separately. Only fits alpha and sigma, not
                                   distribution_weights
        :param kwargs: passed to ``scipy.optimize.minimize``
        :return: If simplfied_fitting=False: scipy OptimizeResult
                 If simplified_fitting=True: List of num_classes OptimizeResults, one for each separate dirichlet fit
        """

        # rescale confidences to avoid divergences on sides of simplex and renormalize
        confidences = (
            confidences * (confidences.shape[0] - 1) + 1 / self.num_classes
        ) / confidences.shape[0]
        confidences = confidences / np.sum(confidences, axis=1)[:, None]

        if not simplified_fitting:
            if initial_parameters is None:
                initial_parameters = np.array([1, 1, 1 / self.num_classes])
            if parameter_bounds is None:
                # dirichlet distribution undefined for alpha/sigma parameters exactly = 0
                parameter_bounds = [(0.0001, None)] * 2 + [(0, 1)]

            # scipy requires an initial guess and a bound (lower, upper) for each parameter
            # not just each parameter class
            initial_parameters = np.repeat(initial_parameters, self.num_classes)
            parameter_bounds = [
                pair for pair in parameter_bounds for i in range(self.num_classes)
            ]

            nll = lambda parms: -np.sum(
                np.log(self.pdf(confidences, *np.split(parms, 3)))
            )
            mle_fit = scipy.optimize.minimize(
                nll, initial_parameters, bounds=parameter_bounds
            )
            self.set_parameters(*np.split(mle_fit.x, 3))

            return mle_fit

        if simplified_fitting:
            if initial_parameters is None:
                initial_parameters = np.array([1, 1])
            if parameter_bounds is None:
                # dirichlet distribution undefined for alpha/sigma parameters exactly = 0
                parameter_bounds = [(0.0001, None)] * 2

            predicted_class = np.argmax(confidences, axis=1)
            class_split_confidences = [
                confidences[predicted_class == i, :] for i in range(self.num_classes)
            ]

            estimated_distribution_weights = [
                k_class_conf.shape[0] for k_class_conf in class_split_confidences
            ]
            estimated_distribution_weights = estimated_distribution_weights / np.sum(
                estimated_distribution_weights
            )

            mle_fits = []

            for k, k_class_confidences in enumerate(class_split_confidences):

                def k_dir_nll(alpha_k, sigma_k):
                    alpha = np.ones(self.num_classes)
                    alpha[k] = alpha_k
                    sigma = np.ones(self.num_classes)
                    sigma[k] = sigma_k
                    # 'isolate' the k'th dirichlet distribution
                    distribution_weights = np.zeros(self.num_classes)
                    distribution_weights[k] = 1
                    return -np.sum(
                        np.log(
                            self.pdf(
                                k_class_confidences, alpha, sigma, distribution_weights
                            )
                        )
                    )

                k_initial_parameters = initial_parameters
                k_parameter_bounds = parameter_bounds

                k_dir_mle_fit = scipy.optimize.minimize(
                    lambda parms: k_dir_nll(*parms),
                    k_initial_parameters,
                    bounds=k_parameter_bounds,
                    **kwargs,
                )
                mle_fits.append(k_dir_mle_fit)

            self.set_alpha(np.array([k_mle_fit.x[0] for k_mle_fit in mle_fits]))
            self.set_sigma(np.array([k_mle_fit.x[1] for k_mle_fit in mle_fits]))
            self.set_distribution_weights(estimated_distribution_weights)

        return mle_fits



import numpy as np

from kyle.models import CalibratableModel


class ModelCalibrator:
    def __init__(
        self,
        X_calibrate: np.ndarray,
        y_calibrate: np.ndarray,
        X_fit: np.ndarray = None,
        y_fit: np.ndarray = None,
    ):
        self.X_calibrate = X_calibrate
        self.y_calibrate = y_calibrate
        self.X_fit = X_fit
        self.y_fit = y_fit

    def calibrate(self, calibratable_model: CalibratableModel, fit: bool = False):
        if fit:
            if self.X_fit is None or self.y_fit is None:
                raise AttributeError("No dataset for fitting provided")
            calibratable_model.fit(self.X_fit, self.y_fit)

        calibratable_model.calibrate(self.X_calibrate, self.y_calibrate)

    def __str__(self):
        return self.__class__.__name__



import numpy as np
from scipy.integrate import quad
from scipy.stats import dirichlet

from kyle.integrals import (
    dirichlet_exp_value,
    simplex_integral_fixed_comp,
    simplex_integral_fixed_max,
)
from kyle.sampling.fake_clf import DirichletFC
from kyle.transformations import SimplexAut


def _prob_correct_prediction(conf: np.ndarray, simplex_aut: SimplexAut):
    conf = conf.squeeze()
    gt_probabilities = simplex_aut.transform(conf, check_io=False)
    return gt_probabilities[np.argmax(conf)]


def _prob_class(conf: np.ndarray, simplex_aut: SimplexAut, selected_class: int):
    conf = conf.squeeze()
    gt_probabilities = simplex_aut.transform(conf, check_io=False)
    return gt_probabilities[selected_class]


def _probability_vector(*parametrization: float):
    return np.array(list(parametrization) + [1 - np.sum(parametrization)])


def compute_accuracy(dirichlet_fc: DirichletFC, **kwargs):
    def integrand(*parametrization):
        conf = _probability_vector(*parametrization)
        return _prob_correct_prediction(conf, dirichlet_fc.simplex_automorphism)

    return dirichlet_exp_value(integrand, dirichlet_fc.alpha, **kwargs)


def compute_expected_max(dirichlet_fc: DirichletFC, **kwargs):
    def integrand(*parametrization):
        conf = _probability_vector(*parametrization)
        return np.max(conf)

    return dirichlet_exp_value(integrand, dirichlet_fc.alpha, **kwargs)


def compute_ECE(dirichlet_fc: DirichletFC, conditioned="full", **kwargs):
    """
    Computes theoretical ECE of dirichlet_fc Fake Classifier conditioned on full confidence vector, conditioned on the
    confidence in prediction or conditioned on each class confidence separately (see [1]_ for further details)

    :param dirichlet_fc: Dirichlet fake classifier to calculate ECE for
    :param conditioned: Quantity to condition ECE on
    :param kwargs: passed to integrator function
    :return: * If conditioned on full confidence vector returns: result, abserr, (further scipy.nquad output)
             * If conditioned on the confidence in prediction returns: result, abserr, (further scipy.quad output)
             * If conditioned on each class separately returns: List of num_classes+1 entries. First entry contains
               average of all "i-class ECEs". Subsequent entries contain results for each "i-class ECE"
               separately: result, abserr, (further scipy.quad output)

    References
    ----------
    .. [1] Kull, M., Perello-Nieto, M., Kängsepp, M., Filho, T. S., Song, H., & Flach, P. (2019). Beyond temperature
        scaling: Obtaining well-calibrated multiclass probabilities with Dirichlet calibration.
    """

    if conditioned == "full":
        return _compute_ECE_full(dirichlet_fc, **kwargs)
    elif conditioned == "confidence":
        return _compute_ECE_conf(dirichlet_fc, **kwargs)
    elif conditioned == "class":
        return _compute_ECE_class(dirichlet_fc, **kwargs)
    else:
        raise ValueError("ECE has to be one of fully, confidence or class conditioned")


def _compute_ECE_full(dirichlet_fc: DirichletFC, **kwargs):
    def integrand(*parametrization):
        conf = _probability_vector(*parametrization)
        return np.abs(
            np.max(conf)
            - _prob_correct_prediction(conf, dirichlet_fc.simplex_automorphism)
        )

    return dirichlet_exp_value(integrand, dirichlet_fc.alpha, **kwargs)


def _compute_ECE_conf(dirichlet_fc: DirichletFC, **kwargs):
    # Need higher precision for accurate result due to nesting of two quad/nquad calls
    # Sets higher precision if precision not already set in **kwargs
    opts = {"epsabs": 1e-4}
    opts.update(kwargs.pop("opts", {}))
    kwargs.update({"opts": opts})

    num_classes = len(dirichlet_fc.alpha)

    def p_c(*parametrization):
        return dirichlet.pdf(parametrization, dirichlet_fc.alpha)

    def p_y_c(*parametrization):
        conf = _probability_vector(*parametrization)
        return _prob_correct_prediction(conf, dirichlet_fc.simplex_automorphism) * p_c(
            *parametrization
        )

    def integrand(max_conf):
        int_p_c = simplex_integral_fixed_max(p_c, num_classes, max_conf, **kwargs)[0]
        int_p_y_c = simplex_integral_fixed_max(p_y_c, num_classes, max_conf, **kwargs)[
            0
        ]
        return np.abs(int_p_y_c / int_p_c - max_conf) * int_p_c

    # At exactly 1/num_classes or 1 get 0/0
    boundary_offset = 1e-2

    return quad(
        integrand,
        1 / num_classes + boundary_offset,
        1 - boundary_offset,
        epsabs=opts["epsabs"],
    )


def _compute_ECE_class(dirichlet_fc: DirichletFC, **kwargs):
    # Need higher precision for accurate result due to nesting of two quad/nquad calls
    # Sets higher precision if precision not already set in **kwargs
    opts = {"epsabs": 1e-4}
    opts.update(kwargs.pop("opts", {}))
    kwargs.update({"opts": opts})

    num_classes = len(dirichlet_fc.alpha)

    integral_results = []

    for i in range(num_classes):

        def p_c(*parametrization):
            return dirichlet.pdf(parametrization, dirichlet_fc.alpha)

        def p_y_c(*parametrization):
            conf = _probability_vector(*parametrization)
            return _prob_class(conf, dirichlet_fc.simplex_automorphism, i) * p_c(
                *parametrization
            )

        def integrand(comp_conf):
            int_p_c = simplex_integral_fixed_comp(
                p_c, num_classes, i, comp_conf, **kwargs
            )[0]
            int_p_y_c = simplex_integral_fixed_comp(
                p_y_c, num_classes, i, comp_conf, **kwargs
            )[0]
            return np.abs(int_p_y_c / int_p_c - comp_conf) * int_p_c

        # At exactly 0 or 1 get 0/0
        boundary_offset = 1e-2

        result = quad(
            integrand,
            1 / num_classes + boundary_offset,
            1 - boundary_offset,
            epsabs=opts["epsabs"],
        )

        integral_results.append(result)

    integral_results.insert(0, sum(S[0] for S in integral_results) / num_classes)

    return integral_results



from typing import Protocol

import numpy as np

from kyle.calibration.calibration_methods import (
    BaseCalibrationMethod,
    TemperatureScaling,
)


class ClassifierProtocol(Protocol):
    def fit(self, X: np.ndarray, y: np.ndarray):
        ...

    def predict(self, X: np.ndarray) -> np.ndarray:
        ...

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        ...


class CalibratableModel(ClassifierProtocol):
    def __init__(
        self,
        model: ClassifierProtocol,
        calibration_method: BaseCalibrationMethod = TemperatureScaling(),
    ):
        self.model = model
        self.calibration_method = calibration_method

    def calibrate(self, X: np.ndarray, y: np.ndarray):
        uncalibrated_confidences = self.model.predict_proba(X)
        self.calibration_method.fit(uncalibrated_confidences, y)

    def fit(self, X: np.ndarray, y: np.ndarray):
        self.model.fit(X, y)

    def predict(self, X: np.ndarray) -> np.ndarray:
        calibrated_proba = self.predict_proba(X)

        return np.argmax(calibrated_proba, axis=2)

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        uncalibrated_confidences = self.model.predict_proba(X)
        return self.calibration_method.get_calibrated_confidences(
            uncalibrated_confidences
        )

    def __str__(self):
        return f"{self.__class__.__name__}, method: {self.calibration_method}"



import logging
from typing import Callable, Protocol, Sequence

from scipy.integrate import nquad
from scipy.stats import dirichlet


# this is the currently supported way to annotate callables with *args of a certain type,
# see https://mypy.readthedocs.io/en/latest/protocols.html#callback-protocols
# hopefully at some point the pycharm type checker will learn to recognize those.
# I opened an issue for JetBrains: https://youtrack.jetbrains.com/issue/PY-45438
class Integrand(Protocol):
    def __call__(self, *parameters: float) -> float:
        ...


def simplex_integral(
    f: Callable, num_classes: int, boundary_offset=1e-10, coord_sum: float = 1, **kwargs
):
    """
    Performs an integral over num_classes-1 dimensional simplex using scipy

    :param f: function to integrate over the simplex. Should accept num_classes-1 variables
    :param num_classes: equals dimension of the simplex + 1
    :param boundary_offset: can be used to prevent numerical errors due to singularities at the simplex' boundary
    :param coord_sum: sets sum of coordinates of simplex. For standard simplex sum(x1,x2,...) = 1. Mainly useful for
        simplex_integral_fixed_max
    :param kwargs: will be passed to scipy.integrate.nquad
    :return:
    """
    if num_classes < 2:
        raise ValueError("need at least two classes")

    def nested_variable_boundary(*previous_variables: float):
        """
        Any variable for the simplex integral goes from zero to coord_sum (usually 1) - sum(all previous variables).
        See docu of nquad for more details on boundaries
        """
        return [
            0 + boundary_offset,
            coord_sum - sum(previous_variables) - boundary_offset,
        ]

    simplex_boundary = [nested_variable_boundary] * (num_classes - 1)
    # we typically don't need higher precision
    opts = {"epsabs": 1e-2}
    opts.update(kwargs.pop("opts", {}))
    return nquad(f, simplex_boundary, opts=opts, **kwargs)


def simplex_integral_fixed_comp(
    f: Callable, num_classes: int, selected_class: int, x_comp: float, **kwargs
):
    """
    Performs an integral over the subset of a num_classes-1 dimensional simplex defined by the selected_class component
    of the confidence vector having a fixed value of x_comp, i.e. marginalises out all other classes.

    Computing this involves integrating over a num_classes-2 dimensional non-unit simplex with coord_sum set to 1-x_comp
    and with the selected_class argument of f being set to x_comp

    :param f: function to integrate over the subset of the simplex. Should accept num_classes-1 variables
    :param num_classes: equals dimension of the simplex + 1
    :param selected_class: selected confidence vector component [0, num_classes-1]
    :param x_comp: fixed value of the selected vector component
    :param kwargs: passed to simplex_integral
    :return:
    """

    if not (0 <= x_comp <= 1):
        raise ValueError("Confidences have to lie in range (0,1)")

    if selected_class == num_classes - 1:

        def constrained_integrand(*args: float):
            constrained_args = [1 - x_comp - sum(args[0:]), *args[0:]]
            return f(*constrained_args)

    else:

        def constrained_integrand(*args: float):
            constrained_args = [*args[0:selected_class], x_comp, *args[selected_class:]]
            return f(*constrained_args)

    return simplex_integral(
        constrained_integrand, num_classes - 1, coord_sum=1 - x_comp, **kwargs
    )


def simplex_integral_fixed_max(f: Callable, num_classes: int, x_max: float, **kwargs):
    """
    Performs an integral over the subset of a num_classes-1 dimensional simplex defined by the largest
    coordinate/confidence having a fixed value of x_max, i.e. marginalises over all possible confidence vectors with
    maximum confidence of x_max.

    Computing this integral involves computing the sum of num_classes integrals each over a num_classes-2 dimensional
    simplex. For x_max > 0.5 the integrals are 'true' simplex integrals. For x_max < 0.5 the boundaries become complex
    and non-simplex like. The integrals can then be extended to full simplex integrals using an appropiate indicator
    function, ``get_argmax_region_char_function``.

    :param f: function to integrate over the subset of the simplex. Should accept num_classes-1 variables
    :param num_classes: equals dimension of the simplex + 1
    :param x_max: fixed value of largest coordinate value. defines subset of simplex
    :param kwargs: passed to simplex_integral_fixed_comp
    :return:
    """

    if not (1 / num_classes < x_max < 1):
        return 0, 0

    # For small x_max higher precision is required for accurate results as over large ingtegration range integrand is 0
    # Sets higher precision if precision not already set in **kwargs
    if x_max < 1 / 2:
        opts = {"epsabs": 1e-4}
        opts.update(kwargs.pop("opts", {}))
        kwargs.update({"opts": opts})

    integral_result = (0, 0)

    for i in range(num_classes):

        argmax_char_func = get_argmax_region_char_function(i)

        constrained_integral = simplex_integral_fixed_comp(
            lambda *args: argmax_char_func(*args) * f(*args),
            num_classes,
            i,
            x_max,
            **kwargs,
        )
        integral_result = tuple(
            sum(p) for p in zip(integral_result, constrained_integral)
        )

    return integral_result


def dirichlet_exp_value(f: Callable, alpha: Sequence[float], **kwargs):
    """
    Computes expectation value of f over num_classes-1 dimensional simplex using scipy. Note scipy.dirichlet.pdf for
    n classes accepts n-1 entries as sum(x_n) = 1.

    :param f:
    :param alpha: the parameters of the dirichlet distribution, one for each class
    :param kwargs: passed to simplex_integral
    :return:
    """
    num_classes = len(alpha)
    return simplex_integral(
        lambda *args: f(*args) * dirichlet.pdf(args, alpha), num_classes, **kwargs
    )


def get_argmax_region_char_function(selected_class: int) -> Integrand:
    """
    Returns the char. function for the area in which the selected class is the argmax of the input args.
    The returned function takes a variable number of floats as input. They represent the first N-1 independent
    entries of an element of a simplex in N-dimensional space (N classes).
    """

    def char_function(*args: float):
        if len(args) < 1:
            raise ValueError("need at least two classes/one input")
        if not 0 <= selected_class <= len(args):
            raise IndexError(
                f"selected_class {selected_class} out of bound for input of length {len(args)}"
            )
        probabilities = list(args) + [1 - sum(args)]
        class_confidence = probabilities[selected_class]
        return float(class_confidence == max(probabilities))

    return char_function


log = logging.getLogger(__name__)



import torch
from kornia.enhance import denormalize
from torch import Tensor, tensor
from torch.utils.data import DataLoader
from torchvision.datasets import CIFAR10
from torchvision.transforms import transforms

# see https://github.com/akamaster/pytorch_resnet_cifar10
resnet_normalize_transform = transforms.Normalize(
    mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
)


def resnet_denormalize_transform(data: Tensor):
    is_batch = len(data.shape) == 4
    if not is_batch:
        data = data[None, :]  # transform only works on batches
    result = denormalize(
        data,
        tensor(resnet_normalize_transform.mean),
        tensor(resnet_normalize_transform.std),
    )
    if not is_batch:
        result = result[0]
    return result


def get_cifar10_dataloader(path: str, train=False):
    dataset = CIFAR10(
        path,
        train=train,
        download=True,
        transform=transforms.Compose(
            [transforms.ToTensor(), resnet_normalize_transform]
        ),
    )
    dataloader = DataLoader(dataset, batch_size=4, shuffle=True, num_workers=1)
    return dataloader


def get_cifar10_dataset(path: str, train=False):
    dataset = CIFAR10(
        path,
        train=train,
        download=True,
        transform=transforms.Compose(
            [transforms.ToTensor(), resnet_normalize_transform]
        ),
    )
    images = []
    targets = []
    # Quick hack, can't find a nice way of doing that. Datasets cannot be sliced and we need the transform
    # Alternative is to retrieve the .data array and do the transformations and reshaping ourselves but this is brittle
    for image, target in dataset:
        images.append(image)
        targets.append(target)
    return torch.stack(images), torch.tensor(targets)



from abc import ABC, abstractmethod

import netcal.metrics
import numpy as np

from kyle.util import in_simplex


# TODO: replace by our own implementations used in EvalStats
class BaseCalibrationError(ABC):
    @abstractmethod
    def _compute(
        self, confidences: np.ndarray, ground_truth: np.ndarray, **kwargs
    ) -> float:
        pass

    def compute(self, confidences: np.ndarray, ground_truth: np.ndarray, **kwargs):
        if not in_simplex(confidences):
            raise ValueError("Invalid confidences array")
        return self._compute(confidences, ground_truth, **kwargs)

    def __str__(self):
        return self.__class__.__name__


class NetcalCalibrationError(BaseCalibrationError):
    def __init__(self, netcal_metric):
        """
        Instance of a netcal metric class, e.g. netcal.metrics.ECE
        """
        self.netcal_metric = netcal_metric

    def _compute(
        self, confidences: np.ndarray, ground_truth: np.ndarray, **kwargs
    ) -> float:
        return self.netcal_metric.measure(confidences, ground_truth, **kwargs)


class ACE(NetcalCalibrationError):
    """Average Calibration Error. Wraps around netcal's implementation - for further reading refer to netcal's docs."""

    def __init__(self, bins: int = 10):
        super(ACE, self).__init__(netcal.metrics.ACE(bins))


class ECE(NetcalCalibrationError):
    """Expected Calibration Error. Wraps around netcal's implementation - for further reading refer to netcal's docs."""

    def __init__(self, bins: int = 10):
        super().__init__(netcal.metrics.ECE(bins))


class MCE(NetcalCalibrationError):
    """Maximum Calibration Error. Wraps around netcal's implementation - for further reading refer to netcal's docs."""

    def __init__(self, bins: int = 10):
        super().__init__(netcal.metrics.MCE(bins))


# TODO: get rid of this



import logging
from typing import Dict, Literal, Sequence, Union

import matplotlib.pyplot as plt
import numpy as np

from kyle.evaluation.reliabilities import (
    _assert_1d,
    _binary_classifier_reliability,
    _plot_reliability_curves,
    _to2d,
    classifier_reliability,
)
from kyle.util import safe_accuracy_score

log = logging.getLogger(__name__)


class EvalStats:
    """
    Class for computing evaluation statistics of classifiers, including calibration metrics
    """

    def __init__(
        self,
        y_true: np.ndarray,
        confidences: np.ndarray,
    ):
        """

        :param y_true: integer array of shape (n_samples,). Assumed to contain true labels in range [0, n_classes-1]
        :param confidences: array of shape (n_samples, n_classes)
        """
        _assert_1d(y_true, "y_true")
        self.y_true = y_true
        self.confidences = _to2d(confidences)

        # Saving some fields, so they don't have to be recomputed
        self.num_samples = len(y_true)
        self.num_classes = confidences.shape[1]
        self.y_pred = confidences.argmax(axis=1)
        # noinspection PyArgumentList
        self._top_class_confidences = np.take_along_axis(
            confidences, self.y_pred[:, None], axis=1
        ).T[0]
        self._argmax_predicted_mask = self.y_true == self.y_pred
        # reshaping b/c the mask can be computed through numpy broadcasting
        possible_labels = np.arange(self.num_classes).reshape((self.num_classes, 1))
        # from this field we can get a mask for whether the label i was predicted
        # by calling self._label_predicted_masks[i]
        # noinspection PyTypeChecker
        self._label_predicted_masks: np.ndarray = self.y_true == possible_labels

    @property
    def top_class_confidences(self):
        return self._top_class_confidences

    def expected_confidence(
        self, class_label: Union[int, Literal["top_class"]] = "top_class"
    ):
        """
        Returns the expected confidence for the selected class or for the predictions (default)

        :param class_label: either the class label as int or "top_class"
        :return:
        """
        if class_label == "top_class":
            confs = self._top_class_confidences
        else:
            confs = self.confidences[:, class_label]
        return float(np.mean(confs))

    def accuracy(self):
        return safe_accuracy_score(self.y_true, self.y_pred)

    def expected_calibration_error(
        self,
        class_label: Union[int, Literal["top_class"]] = "top_class",
        n_bins=12,
        strategy: Literal["uniform", "quantile"] = "uniform",
    ):
        """
        :param class_label: if "top_class", will be the usual (confidence) ECE. Otherwise, it will be the
            marginal class-wise ECE for the selected class.
        :param n_bins:
        :param strategy:
        :return:
        """
        reliabilities = self.reliabilities(
            class_label=class_label,
            n_bins=n_bins,
            strategy=strategy,
        )
        sum_members = np.sum(reliabilities.n_members)
        if sum_members == 0:
            return 0.0
        weights = reliabilities.n_members / sum_members
        abs_diff = np.abs(reliabilities.prob_pred - reliabilities.prob_true)
        return np.dot(abs_diff, weights)

    def average_calibration_error(
        self, n_bins=12, strategy: Literal["uniform", "quantile"] = "uniform"
    ):
        reliabilities = self.reliabilities(
            class_label="top_class", n_bins=n_bins, strategy=strategy
        )
        abs_distances = np.abs(reliabilities.prob_pred - reliabilities.prob_true)
        return np.mean(abs_distances)

    def max_calibration_error(
        self, n_bins=12, strategy: Literal["uniform", "quantile"] = "uniform"
    ):
        reliabilities = self.reliabilities(
            class_label="top_class", n_bins=n_bins, strategy=strategy
        )
        abs_distances = np.abs(reliabilities.prob_pred - reliabilities.prob_true)
        return np.max(abs_distances)

    def class_wise_expected_calibration_error(
        self, n_bins=12, strategy: Literal["uniform", "quantile"] = "uniform"
    ):
        sum_marginal_errors = sum(
            self.expected_calibration_error(k, n_bins=n_bins, strategy=strategy)
            for k in range(self.num_classes)
        )
        return sum_marginal_errors / self.num_classes

    # TODO or not TODO: could in principle work for any 1-dim. reduction but we might not need this generality
    def reliabilities(
        self,
        class_label: Union[int, Literal["top_class"]],
        n_bins=12,
        strategy: Literal["uniform", "quantile"] = "uniform",
    ):
        """
        Computes arrays related to the reliabilities of the provided confidences. They can be used e.g. for computing
        calibration errors or for visualizing reliability curves.

        :param n_bins:
        :param class_label: either an integer label for the class-wise reliabilities, or "top_class" for the
            reliabilities in predictions.
        :param strategy:

        :return: named tuple containing arrays with confidences, accuracies, members, bin_edges
        """
        # Reducing here to save time on re-computation
        if class_label == "top_class":
            reduced_y_pred = self.top_class_confidences
            reduced_y_true = self._argmax_predicted_mask
        else:
            reduced_y_pred = self.confidences[:, class_label]
            reduced_y_true = self._label_predicted_masks[class_label]
        return _binary_classifier_reliability(
            reduced_y_true,
            reduced_y_pred,
            n_bins=n_bins,
            strategy=strategy,
        )

    def plot_reliability_curves(
        self,
        class_labels: Sequence[Union[int, Literal["top_class"]]],
        display_weights=False,
        n_bins=12,
        strategy: Literal["uniform", "quantile"] = "uniform",
    ):
        """

        :param class_labels:
        :param display_weights: If True, for each reliability curve the weights of each bin will be
            plotted as histogram. The weights have been scaled for the sake of display, only relative differences
            between them have an interpretable meaning.
            The errors containing "expected" in the name take these weights into account.
        :param strategy:
        :param n_bins:
        :return: figure
        """
        return _plot_reliability_curves(
            self.reliabilities, class_labels, display_weights, n_bins, strategy
        )

    def plot_gt_distribution(self, label_names: Dict[int, str] = None):
        class_labels, counts = np.unique(self.y_true, return_counts=True)
        if label_names is not None:
            class_labels = [
                label_names.get(label_id, label_id) for label_id in class_labels
            ]

        fig, ax = plt.subplots()
        ax.pie(counts, labels=class_labels, autopct="%1.1f%%", startangle=90)
        ax.axis("equal")  # Equal aspect ratio ensures that pie is drawn as a circle.
        ax.set_title("Ground Truth Distribution")
        return fig



"""
Code taken from: https://github.com/akamaster/pytorch_resnet_cifar10
Proper implementation of ResNet20 for Cifar10. Pytorch only has ResNets for ImageNet which
differ in number of parameters
"""
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init


def _weights_init(m):
    if isinstance(m, nn.Linear) or isinstance(m, nn.Conv2d):
        init.kaiming_normal_(m.weight)


class LambdaLayer(nn.Module):
    def __init__(self, lambd):
        super(LambdaLayer, self).__init__()
        self.lambd = lambd

    def forward(self, x):
        return self.lambd(x)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1, option="A"):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(
            in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False
        )
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(
            planes, planes, kernel_size=3, stride=1, padding=1, bias=False
        )
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != planes:
            if option == "A":
                """
                For CIFAR10 ResNet paper uses option A.
                """
                self.shortcut = LambdaLayer(
                    lambda x: F.pad(
                        x[:, :, ::2, ::2],
                        (0, 0, 0, 0, planes // 4, planes // 4),
                        "constant",
                        0,
                    )
                )
            elif option == "B":
                self.shortcut = nn.Sequential(
                    nn.Conv2d(
                        in_planes,
                        self.expansion * planes,
                        kernel_size=1,
                        stride=stride,
                        bias=False,
                    ),
                    nn.BatchNorm2d(self.expansion * planes),
                )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class ResNet(nn.Module):
    def __init__(self, block, num_blocks, num_classes=10):
        super(ResNet, self).__init__()
        self.in_planes = 16

        self.conv1 = nn.Conv2d(3, 16, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(16)
        self.layer1 = self._make_layer(block, 16, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 32, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 64, num_blocks[2], stride=2)
        self.linear = nn.Linear(64, num_classes)

        self.apply(_weights_init)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion

        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = F.avg_pool2d(out, out.size()[3])
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out


def resnet20():
    return ResNet(BasicBlock, [3, 3, 3])


def resnet56():
    return ResNet(BasicBlock, [9, 9, 9])


def load_weights(weights_path: str, model: ResNet):
    weights_dict = torch.load(weights_path, map_location=torch.device("cpu"))[
        "state_dict"
    ]
    weights_dict = {
        key.replace("module.", ""): value for key, value in weights_dict.items()
    }
    model.load_state_dict(weights_dict)



from abc import ABC, abstractmethod
from typing import Callable, Sequence

import numpy as np
from scipy.special import softmax

from kyle.util import in_simplex


class SimplexAut(ABC):
    """
    Base class for all simplex automorphisms

    :param num_classes: The dimension of the simplex vector, equals 1 + (dimension of the simplex as manifold).
        If provided, will use this for addition I/O checks.
    """

    def __init__(self, num_classes: int = None):
        #  Several transformations can be defined without referring to num_classes, which is why it is optional.
        self.num_classes = num_classes

    def __str__(self):
        return self.__class__.__name__

    @abstractmethod
    def _transform(self, x: np.ndarray) -> np.ndarray:
        """
        :param x: array of shape (n_samples, n_classes)
        :return: transformed array of shape (n_samples, n_classes)
        """
        pass

    def transform(self, x: np.ndarray, check_io=True) -> np.ndarray:
        if len(x.shape) == 1:
            x = x[None, :]
        if check_io and not in_simplex(x, self.num_classes):
            raise ValueError(f"Input has to be from a simplex of suitable dimension")
        x = self._transform(x)
        if check_io and not in_simplex(x, self.num_classes):
            raise ValueError(
                f"Bad implementation: Output has to be from a simplex of suitable dimension"
            )
        return x.squeeze()


class LogitsBasedSimplexAut(SimplexAut, ABC):
    @abstractmethod
    def transform_logits(self, logits: np.ndarray) -> np.ndarray:
        pass

    def _transform(self, x: np.ndarray) -> np.ndarray:
        logits = np.log(x)
        return softmax(self.transform_logits(logits), axis=1)


class TempScaling(LogitsBasedSimplexAut):
    def __init__(self, temperature: float):
        self.temperature = temperature
        super().__init__()

    def transform_logits(self, logits: np.ndarray) -> np.ndarray:
        return self.temperature * logits


class IdentitySimplexAut(SimplexAut):
    def _transform(self, x: np.ndarray) -> np.ndarray:
        return x


class SingleComponentSimplexAut(SimplexAut):
    """
    A simplex automorphism resulting from the application of a map on the unit interval to a
    single component of x and normalizing the result.

    :param component: integer in range [0, num_classes - 1], corresponding to the component on which to apply the mapping
    :param mapping: map from the unit interval [0,1] to itself, should be applicable to arrays
    :param num_classes: The dimension of the simplex vector, equals 1 + (dimension of the simplex as manifold).
        If provided, will use this for addition I/O checks.
    """

    def __init__(
        self,
        component: int,
        mapping: Callable[[np.ndarray], np.ndarray],
        num_classes: int = None,
    ):
        assert (
            0 <= component < num_classes
        ), "Selected component should be in the range [0, num_classes - 1]"
        self.component = component
        self.mapping = mapping
        super().__init__(num_classes=num_classes)

    def _transform(self, x: np.ndarray) -> np.ndarray:
        x = x.copy()
        x[:, self.component] = self.mapping(x[:, self.component])
        return x / x.sum(axis=1)[:, None]


class MaxComponentSimplexAut(SimplexAut):
    """
    A simplex automorphism resulting from the application of a map on the unit interval to a
    the argmax of x and normalizing the remaining components such that the output vector sums to 1.

    :param mapping: map from the unit interval [0,1] to itself, must be applicable to arrays
    :param num_classes: The dimension of the simplex vector, equals 1 + (dimension of the simplex as manifold).
        If provided, will use this for addition I/O checks.
    """

    def __init__(self, mapping: Callable[[np.ndarray], np.ndarray], num_classes=None):
        self.mapping = mapping
        super().__init__(num_classes=num_classes)

    def _transform(self, x: np.ndarray) -> np.ndarray:
        # this transform has a singularity if one component exactly equals one, so we add a minor "noise"
        x = x + 1e-10
        x = x / x.sum(axis=1)[:, None]

        argmax = x.argmax(axis=1)
        old_values = np.choose(argmax, x.T)
        new_values = self.mapping(old_values)
        # the result must sum to 1, so we will rescale the remaining entries of the confidence vectors
        remaining_comps_normalization = (1 - new_values) / (1 - old_values)
        new_values_compensated_for_norm = new_values / remaining_comps_normalization
        np.put_along_axis(
            x, argmax[:, None], new_values_compensated_for_norm[:, None], axis=1
        )
        return x * remaining_comps_normalization[:, None]


class PowerLawSimplexAut(SimplexAut):
    """
    An automorphism resulting from taking elementwise powers of the inputs with fixed exponents
    and normalizing the result.

    |
    | *Intuition*:

    If exponents[j] < exponents[i], then the output will be more shifted towards the j-th direction
    than the i-th. If all exponents are equal to some number s, then s>1 means a shift towards the boundary
    of the simplex whereas 0<s<1 means a shift towards the center and s < 0 results in an "antipodal shift".

    :param exponents: sequence of length num_classes
    """

    def __init__(self, exponents: Sequence[float]):
        self.exponents = np.array(exponents)
        super().__init__(len(exponents))

    def _transform(self, x: np.ndarray) -> np.ndarray:
        x = np.float_power(x, self.exponents)
        return x / x.sum(axis=1)[:, None]


class RestrictedPowerSimplexAut(SimplexAut):
    """
    Maybe a bad idea, feels unnatural
    """

    def __init__(self, exponents: np.ndarray):
        """

        :param exponents: numpy array of shape (num_classes - 1, )
        """
        if not np.all(exponents >= 1):
            raise ValueError("Only exponents >= 1 are permitted")
        self.exponents = exponents[None, :]
        super().__init__(len(exponents) + 1)

    def _transform(self, x: np.ndarray) -> np.ndarray:
        x = x.copy()
        x[:, :-1] = np.float_power(x[:, :-1], self.exponents)
        x[:, -1] = 1 - x[:, :-1].sum(axis=1)
        return x / x.sum(axis=1)[:, None]
